# errors.py - custom dbus bluez errors
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

class BluezDBusException(Exception):
    """
    Custom Bluez Dbus Exception class.
    """
    def __init__(self, reason):
        """ """
        self.reason = reason

    # __init__
    def __str__(self):
        """ """
        return self.reason
    # repr

class DBusFailedError(BluezDBusException):
    pass

class DBusInvalidArgumentsError(BluezDBusException):
    pass

class DBusNotAuthorizedError(BluezDBusException):
    pass

class DBusOutOfMemoryError(BluezDBusException):
    pass

class DBusNoSuchAdapterError(BluezDBusException):
    pass

class DBusNotReadyError(BluezDBusException):
    pass

class DBusNotAvailableError(BluezDBusException):
    pass

class DBusNotConnectedError(BluezDBusException):
    pass

class DBusConnectionAttemptFailedError(BluezDBusException):
    pass

class DBusAlreadyExistsError(BluezDBusException):
    pass

class DBusDoesNotExistError(BluezDBusException):
    pass

class DBusNoReplyError(BluezDBusException):
    pass

class DBusInProgressError(BluezDBusException):
    pass

class DBusNotSupportedError(BluezDBusException):
    pass

class DBusAuthenticationFailedError(BluezDBusException):
    pass

class DBusAuthenticationTimeoutError(BluezDBusException):
    pass

class DBusAuthenticationRejectedError(BluezDBusException):
    pass

class DBusAuthenticationCanceledError(BluezDBusException):
    pass

class DBusUnsupportedMajorClassError(BluezDBusException):
    pass

class DBusNotSupportedError(BluezDBusException):
    pass

class DBusServiceUnknownError(BluezDBusException):
    pass

class DBusMainLoopNotSupportedError(BluezDBusException):
    pass

class DBusMainLoopModuleNotFoundError(BluezDBusException):
    pass

class BluezUnavailableAgentMethodError(BluezDBusException):
    pass


__DICT_ERROR__ = {'org.bluez.Error.Failed:':DBusFailedError,
                  'org.bluez.Error.InvalidArguments:':DBusInvalidArgumentsError,
                  'org.bluez.Error.NotAuthorized:':DBusNotAuthorizedError,
                  'org.bluez.Error.OutOfMemory:':DBusOutOfMemoryError,
                  'org.bluez.Error.NoSuchAdapter:':DBusNoSuchAdapterError,
                  'org.bluez.Error.NotReady:':DBusNotReadyError,
                  'org.bluez.Error.NotAvailable:':DBusNotAvailableError,
                  'org.bluez.Error.NotConnected:':DBusNotConnectedError,
                  'org.bluez.serial.Error.ConnectionAttemptFailed:':DBusConnectionAttemptFailedError,
                  'org.bluez.Error.AlreadyExists:':DBusAlreadyExistsError,
                  'org.bluez.Error.DoesNotExist:':DBusDoesNotExistError,
                  'org.bluez.Error.InProgress:':DBusInProgressError,
                  'org.bluez.Error.NoReply:':DBusNoReplyError,
                  'org.bluez.Error.NotSupported:':DBusNotSupportedError,
                  'org.bluez.Error.AuthenticationFailed:':DBusAuthenticationFailedError,
                  'org.bluez.Error.AuthenticationTimeout:':DBusAuthenticationTimeoutError,
                  'org.bluez.Error.AuthenticationRejected:':DBusAuthenticationRejectedError,
                  'org.bluez.Error.AuthenticationCanceled:':DBusAuthenticationCanceledError,
                  'org.bluez.serial.Error.NotSupported:':DBusNotSupportedError,
                  'org.bluez.Error.UnsupportedMajorClass:':DBusUnsupportedMajorClassError,
                  'org.freedesktop.DBus.Error.ServiceUnknown:':DBusServiceUnknownError}

def parse_dbus_error(exception):
    """
    Return a custom exception based in the dbus exception that was raised.
    """
    global __DICT_ERROR__

    aux = "%s" % exception
    aux_splt = aux.split(None,1)
    try:
        return __DICT_ERROR__[aux_splt[0]](aux_splt[1])
    except KeyError:
        return exception
    # parse_dbus_error



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2009 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.plugins.ConfigPlugin import ConfigPlugin
from blueman.main.SignalTracker import SignalTracker
import cPickle as pickle
import os
import atexit
import dbus.service, dbus.glib
import weakref
import signal
import glib
import copy

def sighandler():
	print "got signal"
	exit()

signal.signal(signal.SIGTERM, sighandler)
signal.signal(signal.SIGHUP, sighandler)

dbus.service.Object.SUPPORTS_MULTIPLE_OBJECT_PATHS = True

cfg_path = os.path.expanduser('~/.config/blueman/blueman.cfg')

class Monitor(dbus.service.Object):
	__id__ = 0
	def __init__(self, plugin):
		Monitor.__id__ += 1
		self.inst_id = "/inst%d" % Monitor.__id__

		self.plugin = plugin
		self.sigs = SignalTracker()
		self.bus = dbus.SessionBus();
		
		dbus.service.Object.__init__(self)
		self.add_to_connection(self.bus, self.inst_id)
		
		self.sigs.Handle("dbus", self.bus, self.on_value_changed, "ValueChanged", "org.blueman.Config")
		
	def on_value_changed(self, section, data):
		s = "".join(chr(b) for b in data)

		(key, value) = pickle.loads(s)
			
		key = str(key)
		
		if self.plugin().section == section:
			self.plugin().set(key, value, True)
		else:
			if not section in File.__db__:
				File.__db__[section] = {}
	
			File.__db__[section][key] = value
		
	@dbus.service.signal(dbus_interface="org.blueman.Config", signature='say')
	def ValueChanged(self, section, data):
		pass

class File(ConfigPlugin):
	__priority__ = 1
	__plugin__ = "file"
	
	__db__ = None
	
	timeout = None

	def __del__(self):
		self.Monitor.sigs.DisconnectAll()
		self.Monitor.remove_from_connection()
		del self.Monitor

	def on_load(self, section):
		
		if not File.__db__:
			if not os.path.exists(os.path.expanduser('~/.config/blueman')):
				try:
					os.makedirs(os.path.expanduser('~/.config/blueman'))
				except:
					pass
			try:
				f = open(cfg_path, "r")
				File.__db__ = pickle.load(f)
				f.close()
			except Exception, e:
				File.__db__ = {}

			atexit.register(File.save)

		
		self.config = File.__db__
		self.section = section
		if self.section == "":
			self.section = "__blueman__"
		
		if not self.section in self.config:
			self.config[self.section] = {}
			
		self.Monitor = Monitor(weakref.ref(self))
	
	@staticmethod	
	def save():
		dprint("Saving config")
		f = open(cfg_path, "w")
		pickle.dump(File.__db__, f, pickle.HIGHEST_PROTOCOL)
		f.close()
		File.timeout = None

	
	def set(self, key, value, local=False):
		
		if key in self.config[self.section]:
			prev = self.config[self.section][key]
		else:
			prev = None
		
		if prev != value:
			self.config[self.section][key] = copy.deepcopy(value)
			self.emit("property-changed", key, value)
			
			if not local:
				if File.timeout:
					glib.source_remove(File.timeout)
				File.timeout = glib.timeout_add(1000, File.save)
				self.Monitor.ValueChanged(self.section, pickle.dumps((key, value), pickle.HIGHEST_PROTOCOL))
		
	def get(self, key):
		if key in self.config[self.section]:
			return copy.deepcopy(self.config[self.section][key])
		else:
			return None
			
	def list_dirs(self):
		l = []
		for key in self.config.iterkeys():
			if self.section in key:
				k = key.replace(self.section, "")
				s = k.split("/")
				if len(s) > 1:
					l.append(self.section + "/" + s[1])
		return l



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

import pynotify
import gtk
from blueman.Functions import dprint
from blueman.gui.GtkAnimation import AnimBase, BezierController

pynotify.init("blueman")

OPACITY_START = 0.7

class Fade(AnimBase):
	def __init__(self, window):
		AnimBase.__init__(self, state=OPACITY_START)
		self.window = window
		
	def state_changed(self, state):
		self.window.props.opacity = state

class NotificationDialog(gtk.MessageDialog):
	def __init__(self, summary, message, timeout=-1, actions=None, actions_cb=None, pixbuf=None, status_icon=None):
		gtk.MessageDialog.__init__(self, parent=None, flags=0, type=gtk.MESSAGE_QUESTION, buttons=gtk.BUTTONS_NONE, message_format=None)
		
		self.bubble = NotificationBubble(summary, message, pixbuf=pixbuf)
		
		i = 100
		self.actions = {}
		self.callback = actions_cb
		if actions:
			for a in actions:
				action_id = a[0]
				action_name = a[1]
				if len(a) == 3:
					icon_name = a[2]
				else:
					icon_name = None
				
				self.actions[i] = action_id
				button = self.add_button(action_name, i)
				if icon_name:
					im = gtk.image_new_from_icon_name(icon_name, gtk.ICON_SIZE_BUTTON)
					im.show()
					button.props.image = im
				i += 1
				
		self.actions[gtk.RESPONSE_DELETE_EVENT] = "close"
		
		self.props.secondary_use_markup = True
		self.resize(350, 50)
		
		self.fader = Fade(self)

		self.props.skip_taskbar_hint = False
		
		self.props.title = summary
		self.props.text = summary
		self.props.secondary_text = message
		
		self.props.window_position = gtk.WIN_POS_CENTER

		if pixbuf:
			self.set_icon_from_pixbuf(pixbuf)
		
		self.connect("response", self.dialog_response)
		self.props.icon_name = "blueman"		
		
		self.entered = False
		def on_enter(widget, event):
			if self.window == gtk.gdk.window_at_pointer()[0] or not self.entered:
				self.fader.animate(start=self.fader.get_state(), end=1.0, duration=500)
				self.entered = True
		
		def on_leave(widget, event):
			if not gtk.gdk.window_at_pointer():
				self.entered = False
				self.fader.animate(start=self.fader.get_state(), end=OPACITY_START, duration=500)
		
		self.connect("enter-notify-event", on_enter)
		self.connect("leave-notify-event", on_leave)

		
		self.set_opacity(OPACITY_START)
		self.present()
		self.set_opacity(OPACITY_START)

		
	def get_id(self):
		if self.bubble:
			return self.bubble.props.id
		
	def dialog_response(self, dialog, response):
		if self.callback:
			self.callback(self, self.actions[response])
		self.hide()
		
	def close(self):
		self.hide()
		
	def set_hint(*args):
		dprint("stub")
		
	def set_timeout(*args):
		dprint("stub")
		
	def add_action(*args):
		dprint("stub")
		
	def clear_actions(*args):
		dprint("stub")
		
	def set_urgency(*args):
		dprint("stub")
		
	def update(self, summary, message):
		self.props.title = summary
		
		self.props.text = summary
		self.props.secondary_text = message
		self.present()
		
	def set_icon_from_pixbuf(self, pixbuf):
		im = gtk.image_new_from_pixbuf(pixbuf)
		self.set_image(im)
		im.show()
	

class NotificationBubble(pynotify.Notification):
	def __init__(self, summary, message, timeout=-1, actions= None, actions_cb=None, pixbuf=None, status_icon=None):
		pynotify.Notification.__init__(self, summary, message)

		def on_notification_closed(n, *args):
			self.disconnect(closed_sig)
			if actions_cb:
				actions_cb(n, "closed")
		
		def on_action(*args):
			self.disconnect(closed_sig)
			actions_cb(*args)	
		
		if pixbuf:
			self.set_icon_from_pixbuf(pixbuf)
		
		if actions:
			for action in actions:
				self.add_action(action[0], action[1], on_action)
			self.add_action("default", "Default Action", on_action)
		
		closed_sig = self.connect("closed", on_notification_closed)
		if timeout != -1:
			self.set_timeout(timeout)
		if status_icon:
			screen, area, orientation = status_icon.get_geometry()
			self.set_hint("x", area.x + area.width/2)
			self.set_hint("y", area.y + area.height/2)
		
		self.show()	
		
	def get_id(self):
		return self.props.id
	

class Notification(object):
	@staticmethod
	def actions_supported():
		return "actions" in pynotify.get_server_caps()
	
	def __new__(cls, summary, message, timeout=-1, actions= None, actions_cb=None, pixbuf=None, status_icon=None):
		if not "actions" in pynotify.get_server_caps():
			if actions != None:
				return NotificationDialog(summary, message, timeout, actions, actions_cb, pixbuf, status_icon)
				
		return NotificationBubble(summary, message, timeout, actions, actions_cb, pixbuf, status_icon)





# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from blueman.gui.Notification import Notification
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.Mechanism import Mechanism
from blueman.main.SignalTracker import SignalTracker
from blueman.Functions import *
import dbus

class DhcpClient(AppletPlugin):
	__description__ = _("Provides a basic dhcp client for Bluetooth PAN connections.")
	__icon__= "network"
	__author__ = "Walmis"
	
	def on_load(self, applet):
		self.Signals = SignalTracker()
		
		self.add_dbus_method(self.DhcpClient, in_signature="s")
			
		self.Signals.Handle("dbus", dbus.SystemBus(), 
						self.on_network_prop_changed, 
						"PropertyChanged",
						"org.bluez.Network",
						path_keyword="path")
						
		self.quering = []
	
	def on_unload(self):
		self.Signals.DisconnectAll()
		
	def DhcpClient(self, interface):
		self.dhcp_acquire(interface)
		
	def on_network_prop_changed(self, key, value, path):
		if key == "Interface":
			if value != "":
				self.dhcp_acquire(value)
		
	def dhcp_acquire(self, device):
		if device not in self.quering:
			self.quering.append(device)
		else:
			return
			
		if device != "":
			def reply(ip_address):

				Notification(_("Bluetooth Network"), _("Interface %(0)s bound to IP address %(1)s") % {"0": device, "1": ip_address}, 
					pixbuf=get_icon("gtk-network", 48), 
					status_icon=self.Applet.Plugins.StatusIcon)
				
				self.quering.remove(device)
			
			def err(msg):
				dprint(msg)
				Notification(_("Bluetooth Network"), _("Failed to obtain an IP address on %s") % (device), 
					pixbuf=get_icon("gtk-network", 48), 
					status_icon=self.Applet.Plugins.StatusIcon)
					
				self.quering.remove(device)
			
			Notification(_("Bluetooth Network"), _("Trying to obtain an IP address on %s\nPlease wait..." % device), 
				pixbuf=get_icon("gtk-network", 48), 
				status_icon=self.Applet.Plugins.StatusIcon)

			m = Mechanism()
			m.DhcpClient(device, reply_handler=reply, error_handler=err, timeout=120)



# utils.py - utils to parse dbus exceptions and check instance type
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus
import errors

def raise_dbus_error(func):
    def warp(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except dbus.DBusException, exception:
            raise errors.parse_dbus_error(exception)
    return warp
# raise_dbus_error

def raise_type_error(instance, cls):
    if not isinstance(instance, cls):
        raise TypeError('Expecting an instance of ' + cls.__name__ + ', not ' + type(instance).__name__)
# raise_type_error



# Adapter.py - class Adapter
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus
import types

import Device
import Agent
import errors
from utils import raise_dbus_error
from utils import raise_type_error
from BaseInterface import BaseInterface

class Adapter(BaseInterface):

    '''Represents the BlueZ dbus API on interface"org.bluez.Adapter".'''

    @raise_dbus_error
    def __init__(self, obj_path):
        super(Adapter, self).__init__('org.bluez.Adapter', obj_path)
    # __init__

    @raise_dbus_error
    def GetProperties(self):
        '''
        Returns all properties for the adapter.
        Properties:
        string Address [readonly]
            The Bluetooth device address.
        string Name [readwrite]
            The Bluetooth friendly name. This value can be
            changed and a PropertyChanged signal will be emitted.
        string Mode [readwrite]
            The Bluetooth operation mode.
            Valid modes: "off", "connectable","discoverable", "limited"
        uint32 DiscoverableTimeout [readwrite]
            The discoverable timeout in seconds. A value of zero
            means that the timeout is disabled and it will stay in
            discoverable/limited mode forever.
            The default value for the discoverable timeout should
            be 180 seconds (3 minutes).
        boolean Discovering [readonly]
            Indicates that a device discovery procedure is active.
        '''
        return self.GetInterface().GetProperties()
    # GetProperties

    @raise_dbus_error
    def SetProperty(self, name, value, **kwargs):
        '''
        Changes the value of the specified property. Only
        properties that are listed a read-write are changeable.
        On success this will emit a PropertyChanged signal.
        '''
        if type(value) == types.IntType:
            value = dbus.UInt32(value)
        self.GetInterface().SetProperty(name, value, **kwargs)
    # SetProperty

    @raise_dbus_error
    def RequestMode(self, mode, reply_handler=None, error_handler=None):
        '''
        This method will request a mode change. The mode
        change must be confirmed by the user via the agent.
        Possible modes for this call are "connectable" and
        "discoverable". Any application that wants to use
        Bluetooth functionality can use this method to
        indicate which mode it needs to operate sucessfully.
        In case the user does not confirm the mode change it
        will return an error to indicate this rejection.
        Use reply_handler and error_handler to make asynchronous call.
        reply_handler will not receive any parameters,
        error_handler will receive an exception as the parameter.
        '''
        def error_handler_wrapper(exception):
            exception = errors.parse_dbus_error(exception)
            if not callable(error_handler):
                raise exception
            error_handler(exception)

        if reply_handler is None and error_handler is None:
            self.GetInterface().RequestMode(mode)
        else:
            self.GetInterface().RequestMode(mode,
                                            reply_handler=reply_handler,
                                            error_handler=error_handler_wrapper)
    # RequestMode

    @raise_dbus_error
    def ReleaseMode(self):
        '''Releases a mode requested via RequestMode.'''
        self.GetInterface().ReleaseMode()
    # ReleaseMode

    @raise_dbus_error
    def StartDiscovery(self):
        '''
        This method starts the device discovery session. This
        includes an inquiry procedure and remote device name
        resolving. Use StopDiscovery to release the sessions
        acquired.
        This process will start emitting DeviceFound and
        PropertyChanged "Discovering" signals.
        '''
        self.GetInterface().StartDiscovery()
    # StartDiscovery

    @raise_dbus_error
    def StopDiscovery(self):
        '''
        This method will cancel any previous StartDiscovery
        transaction.
        Note that a discovery procedure is shared between all
        discovery sessions thus calling StopDiscovery will only
        release a single session.
        '''
        self.GetInterface().StopDiscovery()
    # StopDiscovery

    @raise_dbus_error
    def FindDevice(self, address):
        '''
        Returns the Device instance for given address.
        The device object needs to be first created via
        CreateDevice or CreatePairedDevice.
        '''
        obj_path = self.GetInterface().FindDevice(address)
        return Device.Device(obj_path)
    # FindDevice

    @raise_dbus_error
    def ListDevices(self):
        '''Returns list of Device instances.'''
        obj_paths = self.GetInterface().ListDevices()
        devices = []
        for obj_path in obj_paths:
            devices.append(Device.Device(obj_path))
        return devices
    # ListDevices

    @raise_dbus_error
    def CreateDevice(self, address, reply_handler=None, error_handler=None):
        '''
        Creates a new dbus object path for a remote device,
        then returns Device instance. This method will connect to
        the remote device and retrieve all SDP records.
        If the dbus object path for the remote device already exists
        this method will fail.
        '''
        def reply_handler_wrapper(obj_path):
            if not callable(reply_handler):
                return
            reply_handler(Device.Device(obj_path))

        def error_handler_wrapper(exception):
            exception = errors.parse_dbus_error(exception)
            if not callable(error_handler):
                raise exception
            error_handler(exception)

        if reply_handler is None and error_handler is None:
            obj_path = self.GetInterface().CreateDevice(address)
            return Device.Device(obj_path)
        else:
            self.GetInterface().CreateDevice(address,
                                                   reply_handler=reply_handler_wrapper,
                                                   error_handler=error_handler_wrapper)
            return None


    # CreateDevice

    @raise_dbus_error
    def CreatePairedDevice(self, address, agent, capability='', reply_handler=None, error_handler=None):
        '''
        Creates a new object path for a remote device and then
        returns Device instance. This method will connect to
        the remote device and retrieve all SDP records and then
        initiate the pairing.
        If previously CreateDevice was used successfully,
        this method will only initiate the pairing.
        Compared to CreateDevice this method will fail if
        the pairing already exists, but not if the object
        path already has been created. This allows applications
        to use CreateDevice first and the if needed use
        CreatePairedDevice to initiate pairing.
        The capability parameter is the same as for the
        RegisterAgent method.
        Use reply_handler and error_handler to make asynchronous call,
        and this method will return None.
        reply_handler will receive an instance of Device as the parameter,
        error_handler will receive an exception as the parameter.
        '''
        def reply_handler_wrapper(obj_path):
            if not callable(reply_handler):
                return
            reply_handler(Device.Device(obj_path))

        def error_handler_wrapper(exception):
            exception = errors.parse_dbus_error(exception)
            if not callable(error_handler):
                raise exception
            error_handler(exception)

        if reply_handler is None and error_handler is None:
            obj_path = self.GetInterface().CreatePairedDevice(address, agent, capability)
            return Device.Device(obj_path)
        else:
            self.GetInterface().CreatePairedDevice(address,
                                                   agent,
                                                   capability,
                                                   reply_handler=reply_handler_wrapper,
                                                   error_handler=error_handler_wrapper)
            return None
    # CreatePairedDevice

    @raise_dbus_error
    def RemoveDevice(self, device):
        '''
        This removes the remote device dbus object according to
        given Device instance. It will remove also the pairing information.
        '''
        raise_type_error(device, Device.Device)
        self.GetInterface().RemoveDevice(device.GetObjectPath())
    # RemoveDevice

    @raise_dbus_error
    def RegisterAgent(self, agent, capability=''):
        '''
        This registers the adapter wide agent.
        The agent should be an instance of Agent, the methods of Agent
        will be called when user input is needed.
        If an application disconnects from the dbus bus all
        of its registered agents will be removed.
        The capability parameter can have the values
        "DisplayOnly", "DisplayYesNo", "KeyboardOnly" and
        "NoInputNoOutput" which reflects the input and output
        capabilities of the agent. If an empty string is
        used it will fallback to "DisplayYesNo".
        '''
        raise_type_error(agent, Agent.Agent)
        self.GetInterface().RegisterAgent(agent.GetObjectPath(), capability)
    # RegisterAgent

    @raise_dbus_error
    def UnregisterAgent(self, agent):
        '''
        This unregisters the agent that has been previously
        registered.
        '''
        raise_type_error(agent, Agent.Agent)
        self.GetInterface().UnregisterAgent(agent.GetObjectPath())
    # UnregisterAgent
# Adapter



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
import gtk


class GenericList(gtk.TreeView):

	def __init__(self, data):
		gtk.TreeView.__init__(self)
		self.selection = self.get_selection()
		self._load(data)
		
		

	def _load(self, data):
		
		self.ids = {}
		self.columns = {}
		

		types = []
		for i in range(len(data)):
			types.append(data[i][1])
			
		types = tuple(types)
		
		self.liststore = gtk.ListStore(*types)
		self.set_model(self.liststore)
		column = None
		for i in range(len(data)):

			self.ids[data[i][0]] = i
			
			
			if len(data[i]) == 5 or len(data[i]) == 6:

				column = gtk.TreeViewColumn(data[i][4])
				
				column.pack_start(data[i][2])
				column.set_attributes(data[i][2], **data[i][3])				
				
				
				if len(data[i]) == 6:
					column.set_properties(**data[i][5])
				
				self.columns[data[i][0]] = column
				self.append_column(column)
				

    
	def selected(self):
		(model, iter) = self.selection.get_selected()
               
               	return iter

                	
    
    	def delete(self, id):
		if type(id) == gtk.TreeIter:
			iter = id
		else:
			iter = self.get_iter(id)
    		
    		if iter == None:
    			return False
    		if self.liststore.iter_is_valid(iter):
    			self.liststore.remove(iter)
    			return True
    		else:
    			return False
    			
    	
    	def _add(self, **columns):
    		ids_len = len(self.ids)
    		cols_len = len(columns)
    		
		items = {}
    		for k, v in self.ids.iteritems():
    			items[v] = None
	
 		
    		for k, v in columns.iteritems():
    			if k in self.ids:
    				items[self.ids[k]] = v
    			else:
    				raise Exception, "Invalid key %s" % k
    				
		return items.values()
		
	def append(self, **columns):
		vals = self._add(**columns)
		return self.liststore.append(vals)
	
	def prepend(self, **columns):
		vals = self._add(**columns)
		return self.liststore.prepend(vals)
    		
	def get_conditional(self, **cols):
   		ret = []
   		matches = 0
   		for i in range(len(self.liststore)):
   			row = self.get(i)
    			for k, v in cols.iteritems():
    				if row[k] == v:
    					matches += 1
    					
    			if matches == len(cols):
    				ret.append(i)
    				
    			matches = 0

    		return ret
    		
    		
    	def set(self, id, **cols):
		if type(id) == gtk.TreeIter:
			iter = id
		else:
			iter = self.get_iter(id)
		
		if iter != None:
	    		for k, v in cols.iteritems():
	    			self.liststore.set(iter, self.ids[k], v)
	    		return True
		else:
	    		return False

    	
    	
    	def get(self, id, *items):
		ret = {}

		if id != None:
			if type(id) == gtk.TreeIter:
				iter = id
			else:
				iter = self.get_iter(id)
			if len(items) == 0:
				for k, v in self.ids.iteritems():
					ret[k] = self.liststore.get(iter, v)[0]
			else:
				for i in range(len(items)):
					if items[i] in self.ids:
						ret[items[i]] = self.liststore.get(iter, self.ids[items[i]])[0]
		else:
			return False

	
		return ret

        	
  	
    	def get_iter(self, path):
        	if path == None:
        		return None
        	
        	try:
            		return self.liststore.get_iter(path)
        	except:
            		return None
            		
	def clear(self):
		self.liststore.clear()
		
	def compare(self, iter_a, iter_b):
		if iter_a != None and iter_b != None:
			return self.get_model().get_path(iter_a) == self.get_model().get_path(iter_b)
		else:
			return False




# Copyright (C) 2010 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.plugins.BasePlugin import BasePlugin

class ManagerPlugin(BasePlugin):

	#return list of (GtkMenuItem, position) tuples
	def on_request_menu_items(self, manager_menu, device):
		pass



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin

import gobject
import gtk

class Menu(AppletPlugin):
	__depends__ = ["StatusIcon"]
	__description__ = _("Provides a menu for the applet and an API for other plugins to manipulate it")
	__icon__ = "menu-editor"
	__author__ = "Walmis"
	__unloadable__ = False
	
	def on_load(self, applet):
		self.Applet = applet
		
		self.Applet.Plugins.StatusIcon.connect("popup-menu", self.on_popup_menu)
		
		self.__plugins_loaded = False
		
		self.__menuitems = []
		self.__menu = gtk.Menu()
		
	def on_popup_menu(self, status_icon, button, activate_time):
		self.__menu.popup(None, None, gtk.status_icon_position_menu,
						button, activate_time, status_icon)		
	
	def __sort(self):
		self.__menuitems.sort(lambda a, b: cmp(a[0], b[0]))
		
	def __clear(self):
		def each(child):
			self.__menu.remove(child)
		self.__menu.foreach(each)	
	
	def __load_items(self):
		for item in self.__menuitems:
			self.__menu.append(item[1])
			if item[2]:
				item[1].show()		
	
	def Register(self, owner, item, priority, show=True):
		self.__menuitems.append((priority, item, show, owner))
		if self.__plugins_loaded:
			self.__sort()
			self.__clear()
			self.__load_items()
			
		
	def Unregister(self, owner):
		for i in reversed(self.__menuitems):
			priority, item, show, orig_owner = i
			if orig_owner == owner:
				self.__menu.remove(item)
				self.__menuitems.remove(i)
		
	def on_plugins_loaded(self):
		self.__plugins_loaded = True
		self.__sort()
		self.__load_items()
		
	def get_menu(self):
		return self.__menu
		
			




# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
import socket
import fcntl
import gobject
import struct
import subprocess
from blueman.Lib import get_net_address


class DhcpClient(gobject.GObject):
	__gsignals__ = {
		#arg: interface name eg. ppp0
		'connected' : (gobject.SIGNAL_NO_HOOKS, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,)),
		'error-occurred' : (gobject.SIGNAL_NO_HOOKS, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,)),
	}	
	
	quering = []
	
	def __init__(self, interface, timeout=30):
		gobject.GObject.__init__(self)
	
		self.interface = interface
		self.timeout = timeout
	
	def Connect(self):
		if self.interface in DhcpClient.quering:
			raise Exception("dhclient already running on this interface")
		else:
			DhcpClient.quering.append(self.interface)
			
		try:
			self.dhclient = subprocess.Popen(["dhclient", "-e", "IF_METRIC=100", "-1", self.interface])
		except:
			raise Exception("dhclient binary not found")
			
		gobject.timeout_add(1000, self.check_dhclient)
		gobject.timeout_add(self.timeout*1000, self.on_timeout)

		
	def on_timeout(self):
		if self.dhclient.poll() == None:
			dprint("Timeout reached, terminating dhclient")
			self.dhclient.terminate()
		
	def check_dhclient(self):
		status = self.dhclient.poll()
		if status != None:
			if status == 0:
				def complete():
					ip = get_net_address(self.interface)
					dprint("bound to", ip)
					self.emit("connected", ip)	
				
				gobject.timeout_add(1000, complete)
				DhcpClient.quering.remove(self.interface)
				return False
				
			else:
				dprint("dhclient failed with status code", status)
				self.emit("error-occurred", status)
				DhcpClient.quering.remove(self.interface)
				return False
			
		return True		



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluez.Device import Device as BluezDevice
from blueman.main.Device import Device
from blueman.gui.Notification import Notification
from blueman.main.PulseAudioUtils import PulseAudioUtils, EventType
from subprocess import Popen, PIPE
import gobject

import dbus
from blueman.main.SignalTracker import SignalTracker

class SourceRedirector:
	instances = []
	def __init__(self, module_id, device_path, pa_utils):
		if module_id in SourceRedirector.instances:
			return
		else:
			SourceRedirector.instances.append(module_id)
		
		self.module_id = module_id
		self.pa_utils = pa_utils
		self.device = Device(device_path)
		self.signals = SignalTracker()
		self.bus = dbus.SystemBus()
		self.signals.Handle("dbus", self.bus, self.on_source_prop_change, "PropertyChanged", "org.bluez.AudioSource", path=device_path)
		
		self.pacat = None
		self.parec = None
		self.loopback_id = None
		
		dprint("Starting source redirector")
		def sources_cb(sources):
			for k, v in sources.iteritems():
				props = v["proplist"]
				if "bluetooth.protocol" in props:
					if props["bluetooth.protocol"] == "a2dp_source":
						if v["owner_module"] == self.module_id:
							dprint("Found source", k)
							self.start_redirect(k)
							return
			dprint("Source not found :(")
		
		self.pa_utils.ListSources(sources_cb)
		
	def start_redirect(self, source):

		def on_load(res):
			dprint("module-loopback load result", res)
			if res < 0:
				self.parec = Popen(["parec", "-d", str(source)], stdout=PIPE)
				self.pacat = Popen(["pacat", "--client-name=Blueman", "--stream-name=%s" % self.device.Address, "--property=application.icon_name=blueman"], stdin=self.parec.stdout)
			else:
				self.loopback_id = res
				
		self.pa_utils.LoadModule("module-loopback", "source=%d" % source, on_load)
		
	def on_source_prop_change(self, key, value):
		if key == "State":
			if value == "disconnected":
				if self.pacat:
					self.pacat.terminate()
				if self.parec:
					self.parec.terminate()
				if self.loopback_id:
					self.pa_utils.UnloadModule(self.loopback_id, lambda x: dprint("Loopback module unload result", x))
				
				self.signals.DisconnectAll()
				
				SourceRedirector.instances.remove(self.module_id)

				del self.pa_utils
				
	def __del__(self):
		dprint("Destroying redirector")
		
class Module(gobject.GObject):
	__gsignals__ = {
		'loaded' : (gobject.SIGNAL_NO_HOOKS, gobject.TYPE_NONE, ()),
	}	
	def __init__(self):
		gobject.GObject.__init__(self)
		self.refcount = 0
		self.id = None
		
	def unload(self):
		dprint(self.id)
		pa = PulseAudioUtils()
		id = self.id
		pa.UnloadModule(self.id, lambda x: dprint("Unload %s result %s" % (id, x)))
		self.id = None
		self.refcount = 0
		
	def ref(self):
		self.refcount+=1
		
		dprint(self.id, self.refcount)
		
	def unref(self):
		self.refcount-=1
		
		dprint(self.id, self.refcount)
		
		if self.refcount <= 0 and self.id:
			self.unload()
			
	def load(self, args, cb):
		if self.id != None:
			self.unload()
		
		def load_cb(res):
			if res > 0:
				self.refcount = 1
				self.id = res
				if cb:
					cb(res)
				self.emit("loaded")
			else:
				self.refcount = 0
				self.id = None
		
		PulseAudioUtils().LoadModule("module-bluetooth-device",
									 args,
									 load_cb)

class PulseAudio(AppletPlugin):
	__author__ = "Walmis"
	__description__ = _("Automatically manages Pulseaudio Bluetooth sinks/sources.\n"
						"<b>Note:</b> Requires pulseaudio 0.9.15 or higher")
	__icon__ = "audio-card"
	__options__  = {
		"checked" : {"type": bool, "default": False},
		"make_default_sink": {"type":bool, 
							  "default": True, 
							  "name": _("Make default sink"), 
							  "desc": _("Make the a2dp audio sink the default after connection")},
		"move_streams": {"type": bool, 
						 "default": True, 
						 "name": _("Move streams"), 
						 "desc": _("Move existing audio streams to bluetooth device")}
	}
	def on_load(self, applet):
		self.signals = SignalTracker()
		if not self.get_option("checked"):
			self.set_option("checked", True)
			if not have("pactl"):
				applet.Plugins.SetConfig("PulseAudio", False)
				return
			
		self.bus = dbus.SystemBus()
		
		self.connected_sources = []
		self.connected_sinks = []
		self.connected_hs = []
		
		self.loaded_modules = {}
		
		self.pulse_utils = PulseAudioUtils()
		version = self.pulse_utils.GetVersion()
		dprint("PulseAudio version:", version)
		
		if version[0] == 0:
			if tuple(version) < (0, 9, 15):
				raise Exception("PulseAudio too old, required 0.9.15 or higher")
		
		self.signals.Handle("dbus", 
							self.bus, 
							self.on_sink_prop_change, 
							"PropertyChanged", 
							"org.bluez.AudioSink", 
							path_keyword="device")
							
		self.signals.Handle("dbus", 
							self.bus, 
							self.on_source_prop_change, 
							"PropertyChanged", 
							"org.bluez.AudioSource", 
							path_keyword="device")
							
		self.signals.Handle("dbus", 
							self.bus, 
							self.on_hsp_prop_change, 
							"PropertyChanged", 
							"org.bluez.Headset", 
							path_keyword="device")
		

		self.signals.Handle(self.pulse_utils, "connected", self.on_pulse_connected)
		self.signals.Handle(self.pulse_utils, "event", self.on_pulse_event)
	
	def on_pulse_event(self, pa_utils, event, idx):
		if (EventType.CARD | EventType.CHANGE) == event:
			dprint(event)
			def card_cb(c):
				dprint(c)
				m = self.loaded_modules[c["proplist"]["bluez.path"]]
				if c["owner_module"] == m.id:
					if c["active_profile"] == "a2dp_source":
						SourceRedirector(m.id, c["proplist"]["bluez.path"], pa_utils)
			
			pa_utils.GetCard(idx, card_cb)
		
	
	def on_pulse_connected(self, pa_utils):
		def modules_cb(modules):
			for k, v in modules.iteritems():
				if v["name"] == "module-bluetooth-discover":
					pa_utils.UnloadModule(k, lambda x: dprint("Unload module-bluetooth-discover result", x))
			
		self.pulse_utils.ListModules(modules_cb)			
	
	def on_unload(self):
		self.signals.DisconnectAll()
		
	def load_module(self, dev_path, args, cb=None):
		if not dev_path in self.loaded_modules:
			m = Module()
			m.load(args, cb)
			self.loaded_modules[dev_path] = m
		else:
			self.loaded_modules[dev_path].ref()
		
	def try_unload_module(self, dev_path):
		try:
			m = self.loaded_modules[dev_path]
			m.unref()
			if m.refcount == 0:
				del self.loaded_modules[dev_path]
		except Exception, e:
			dprint(e)
		
	def on_source_prop_change(self, key, value, device):
		dprint(key, value)
			
		if key == "State":
			if value == "connected":
				if not device in self.connected_sources:
					self.connected_sources.append(device)
					d = Device(device)
					self.load_module(device, "path=%s address=%s profile=a2dp_source source_properties=device.icon_name=blueman card_properties=device.icon_name=blueman" % (device, d.Address))
					
			elif value == "disconnected":
				self.try_unload_module(device)
				if device in self.connected_sources:
					self.connected_sources.remove(device)

			elif value == "playing":
				try:
					m = self.loaded_modules[device]
					def on_loaded(m):
						SourceRedirector(m.id, device, self.pulse_utils)
						m.disconnect(sig)
					
					if not m.id:
						sig = m.connect("loaded", on_loaded)
					else:
						SourceRedirector(m.id, device, self.pulse_utils)
					
				except Exception, e:
					dprint(e)
				
		
	def on_sink_prop_change(self, key, value, device):
		if key == "Connected" and value:
			if not device in self.connected_sinks:
				self.connected_sinks.append(device)
				gobject.timeout_add(500, self.setup_pa, device, "a2dp")
		
		elif key == "Connected" and not value:
			if device in self.connected_sinks:
				self.connected_sinks.remove(device)
			self.try_unload_module(device)
		
	def on_hsp_prop_change(self, key, value, device):
		if key == "Connected" and value:
			if not device in self.connected_hs:
				self.connected_hs.append(device)
				self.setup_pa(device, "hsp")
		
		elif key == "Connected" and not value:
			self.try_unload_module(device)
			if device in self.connected_hs:
				self.connected_hs.remove(device)
			
	def move_pa_streams(self, sink_id):
		def inputs_cb(inputs):
			for k, v in inputs.iteritems():
				dprint("moving stream", v["name"], "to sink", sink_id)
				self.pulse_utils.MoveSinkInput(k, sink_id, None)
		
		self.pulse_utils.ListSinkInputs(inputs_cb)
				
	def setup_pa_sinks(self, module_id):
		dprint("module", module_id)
		
		def sinks_cb(sinks):
			for k, v in sinks.iteritems():
				if v["owner_module"] == module_id:
					if self.get_option("make_default_sink"):
						dprint("Making sink", v["name"], "the default")
						self.pulse_utils.SetDefaultSink(v["name"], None)
					if self.get_option("move_streams"):
						self.move_pa_streams(k)
					
		
		self.pulse_utils.ListSinks(sinks_cb)
		
	def setup_pa(self, device_path, profile):
		device = Device(device_path)
		
		
		def load_cb(res):
			dprint("Load result", res)

			if res < 0:
				Notification(_("Bluetooth Audio"), 
							 _("Failed to initialize PulseAudio Bluetooth module. Bluetooth audio over PulseAudio will not work."), 
							 pixbuf=get_notification_icon("gtk-dialog-error"), 
							 status_icon=self.Applet.Plugins.StatusIcon)				
			else:
				Notification(_("Bluetooth Audio"), 
							 _("Successfully connected to a Bluetooth audio device. This device will now be available in the PulseAudio mixer"), 
							 pixbuf=get_notification_icon("audio-card"), 
							 status_icon=self.Applet.Plugins.StatusIcon)	
				if profile == "a2dp":			 
					self.setup_pa_sinks(res)
			
			#connect to other services, so pulseaudio profile switcher could work
			for s in ("headset", "audiosink", "audiosource"):	
				try:
					device.Services[s].Connect()
				except KeyError:
					pass
				except Exception, e:
					print e
				
		version = self.pulse_utils.GetVersion()
		if version[0] == 1 or version[2] >= 18:
			args = "address=%s profile=%s sink_properties=device.icon_name=blueman card_properties=device.icon_name=blueman"
		else:
			args = "address=%s profile=%s"

		self.load_module(device_path, args % (device.Address, profile), load_cb)
		
		
		



# ServiceInterface.py - interface of BlueZ service plugins
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from utils import raise_dbus_error
from BlueZInterface import BlueZInterface

class ServiceInterface(BlueZInterface):

    @raise_dbus_error
    def __init__(self, interface, obj_path, methods):
        self.methods = methods
        super(ServiceInterface, self).__init__(interface, obj_path)
    # __init__

    def __getattribute__(self, name):
        if name in object.__getattribute__(self, 'methods'):
            func = getattr(self.GetInterface(), name)
            return raise_dbus_error(func)
        else:
            return super(ServiceInterface, self).__getattribute__(name)
    # __getattribute__
# ServiceInterface



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

import glib
import tty
import termios
import os
import subprocess
import gobject
import errno
import re

pppd_errors = {}
pppd_errors[1] = """An immediately fatal error of some kind  occurred, such as an essential system call failing, or running out of virtual memory."""
pppd_errors[2] = """An  error  was detected in processing the options given, such as two mutually exclusive options being used."""
pppd_errors[3] = """Pppd is not setuid-root and the invoking user is not root."""
pppd_errors[4] = """The kernel does not support PPP, for example, the  PPP kernel driver is not included or cannot be loaded."""
pppd_errors[5] = """Pppd terminated because it was sent a SIGINT, SIGTERM or SIGHUP signal."""
pppd_errors[6] = """The serial port could not be locked."""
pppd_errors[7] = """The serial port could not be opened."""
pppd_errors[8] = """The connect script failed (returned a non-zero exit status)."""
pppd_errors[9] = """The command specified as the argument to the  pty  option  could not be run."""
pppd_errors[10] = """The PPP negotiation failed, that is, it didn't reach the point where at least one network protocol (e.g. IP) was running."""
pppd_errors[11] = """The peer system failed (or refused) to authenticate itself."""
pppd_errors[12] = """The link was established successfully and terminated because  it was idle."""
pppd_errors[13] = """The link was established successfully and terminated because the connect time limit was reached."""
pppd_errors[14] = """Callback was negotiated  and  an  incoming  call  should  arrive shortly."""
pppd_errors[15] = """The link was terminated because the peer is not responding to echo requests."""
pppd_errors[16] = """The link was terminated by the modem hanging up."""
pppd_errors[17] = """The PPP negotiation failed because serial loopback was detected."""
pppd_errors[18] = """The init script failed (returned a non-zero exit status)."""
pppd_errors[19] = """We failed to authenticate ourselves to the peer."""


class PPPException(Exception):
	pass

class PPPConnection(gobject.GObject):
	__gsignals__ = {
		#arg: interface name eg. ppp0
		'connected' : (gobject.SIGNAL_NO_HOOKS, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,)),
		'error-occurred' : (gobject.SIGNAL_NO_HOOKS, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,)),

	}
	def __init__(self, port, number="*99#", apn="", user="", pwd=""):
		gobject.GObject.__init__(self)
		
		self.apn = apn
		self.number = number
		self.user = user
		self.pwd = pwd
		self.port = port
		self.interface = None
		
		self.pppd = None
		self.file = None
		
		self.commands = [("ATZ E0 V1 X4 &C1 +FCLASS=0", self.simple_callback), 
				 ("ATE0", self.simple_callback), 
				 ("AT+GCAP", self.simple_callback), 
				 ("ATD%s" % self.number, self.connect_callback, ["CONNECT", "NO CARRIER", "BUSY", "NO ANSWER", "NO DIALTONE", "OK", "ERROR"])
				]
		if self.apn != "":		
			self.commands.insert(-1, ('AT+CGDCONT=1,"IP","%s"' % self.apn, self.simple_callback))
				
	def cleanup(self):
		os.close(self.file)
		self.file = None
	
	def simple_callback(self, response):
		pass
		
	def connect_callback(self, response):
		if "CONNECT" in response:
			dprint("Starting pppd")
			self.pppd = subprocess.Popen(["/usr/sbin/pppd", "%s" % self.port, "115200", "defaultroute", "updetach", "usepeerdns"], bufsize=1, stdout=subprocess.PIPE)
			glib.io_add_watch(self.pppd.stdout, glib.IO_IN | glib.IO_ERR | glib.IO_HUP, self.on_pppd_stdout)
			glib.timeout_add(1000, self.check_pppd)				
			
			self.cleanup()
		else:
			self.cleanup()
			raise PPPException("Bad modem response %s, expected CONNECT" % response[0])
	
	def __cmd_response_cb(self, response, exception, item_id):
		if exception:
			self.emit("error-occurred", str(exception))
		else:	
			try:
				self.commands[item_id][1](response)
			except PPPException, e:
				self.emit("error-occurred", str(e))
				return
				
			self.send_commands(item_id+1)

	
	def send_commands(self, id=0):
		try:
			item = self.commands[id]
		except IndexError:
			return
			
		if len(item) == 3:
			(command, callback, terminators) = item
		else:
			(command, callback) = item
			terminators = ["OK", "ERROR"]
			
		self.send_command(command)
		self.wait_for_reply(self.__cmd_response_cb, terminators, id)	
				

	def Connect(self):
			
		self.file = os.open(self.port, os.O_RDWR | os.O_EXCL | os.O_NONBLOCK | os.O_NOCTTY)
		
		tty.setraw(self.file)
		
		attrs = termios.tcgetattr(self.file)
		
		attrs[0] &= ~(termios.IGNCR | termios.ICRNL | termios.IUCLC | termios.INPCK | termios.IXON | termios.IXANY | termios.IGNPAR)
		attrs[1] &= ~(termios.OPOST | termios.OLCUC | termios.OCRNL | termios.ONLCR | termios.ONLRET)
		attrs[3] &= ~(termios.ICANON | termios.XCASE | termios.ECHO | termios.ECHOE | termios.ECHONL)
		attrs[3] &= ~(termios.ECHO | termios.ECHOE)
		attrs[6][termios.VMIN] = 1
		attrs[6][termios.VTIME] = 0
		attrs[6][termios.VEOF] = 1

		attrs[2] &= ~(termios.CBAUD | termios.CSIZE | termios.CSTOPB | termios.CLOCAL | termios.PARENB)
		attrs[2] |= (termios.B9600 | termios.CS8 | termios.CREAD | termios.PARENB)
		
		termios.tcsetattr(self.file, termios.TCSANOW, attrs)
		
		termios.tcflush(self.file, termios.TCIOFLUSH)
		
		self.send_commands()	
		
	def on_pppd_stdout(self, source, cond):
		if cond & glib.IO_ERR or cond & glib.IO_HUP:
			return False

		line = source.readline()	
		m = re.match("Using interface (ppp[0-9]*)", line)
		if m:
			self.interface = m.groups(1)[0]
		
		print line
	
		return True
	
	def check_pppd(self):
		status = self.pppd.poll()	
		if status != None:
			if status == 0:
				self.emit("connected", self.interface)
			else:
				try:
					msg = "pppd exited: " + pppd_errors[int(status)]
				except KeyError:
					msg = "pppd exited with unknown error"
					
				self.emit("error-occurred", msg)
				
			print "pppd exited with status %d" % status
			return False
		return True

	def send_command(self, command):
		dprint("-->", command)
		os.write(self.file, "%s\r\n" % command)
		termios.tcdrain(self.file)
		
	def on_data_ready(self, source, condition, terminators, on_done):
		if condition & glib.IO_ERR or condition & glib.IO_HUP:
			on_done(None, PPPException("Socket error"))
			self.cleanup()
			return False
		try:
			self.buffer += os.read(self.file, 1)
		except OSError, e:
			if e.errno == errno.EAGAIN:
				dprint("Got EAGAIN")
				return True
			else:
				on_done(None, PPPException("Socket error"))
				dprint(e)
				self.cleanup()
				return False

		lines = self.buffer.split("\r\n")
		found = False
		for l in lines:
			if l == "":
				pass		
			else:
				for t in terminators:
					if t in l:
						found = True

		if found:
			lines = filter(lambda x: x != "", lines)
			lines = map(lambda x: x.strip("\r\n"), lines)
			dprint("<-- ", lines)
			
			on_done(lines, None)
			return False		
			
		return True
		
	def wait_for_reply(self, callback, terminators=["OK", "ERROR"], *user_data):
		def on_timeout():
			glib.source_remove(self.io_watch)
			callback(None, PPPException("Modem initialization timed out"), *user_data)
			self.cleanup()
			return False
			
			
		def on_done(ret, exception):
			glib.source_remove(self.timeout)
			callback(ret, exception, *user_data)
			
		
		self.buffer = ""
		self.term_found = False
				
		self.io_watch = glib.io_add_watch(self.file, glib.IO_IN | glib.IO_ERR | glib.IO_HUP, self.on_data_ready, terminators, on_done)
		self.timeout = glib.timeout_add(15000, on_timeout)
		



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from blueman.Functions import *
from blueman.main.Config import Config
from blueman.plugins.AppletPlugin import AppletPlugin
import dbus

class Headset(AppletPlugin):
	__author__ = "Walmis"
	__description__ = _("Runs a command when answer button is pressed on a headset")
	__icon__ = "blueman-headset"
	
	__options__  = {
		"command" : {"type": str,
				  "default": "",
				  "name": _("Command"),
				  "desc": _("Command to execute when answer button is pressed:")
				  }
	}
		
	def on_load(self, applet):
		self.bus = dbus.SystemBus()
		self.bus.add_signal_receiver(self.on_answer_requested, "AnswerRequested", "org.bluez.Headset")
		
	def on_unload(self):
		self.bus.remove_signal_receiver(self.on_answer_requested, "AnswerRequested", "org.bluez.Headset")
		
	def on_answer_requested(self):
		c = self.get_option("command")
		if c and c != "":
			args = c.split(" ")
			try:
				spawn(args, True)
			except:
				dprint("Cannot launch command")





VERSION="1.23"
PACKAGE="blueman"
WEBSITE="https://github.com/blueman-project/blueman"
PREFIX="/usr/local"
BIN_DIR="/usr/local/bin"
ICON_PATH = "/usr/local/share/blueman/icons"
UI_PATH = "/usr/local/share/blueman/ui"
OBEX_BROWSE_AVAILABLE = True
DHCP_CONFIG_FILE = "/etc/dhcp3/dhcpd.conf"
POLKIT = "yes" == "yes"
HAL_ENABLED = "no" == "yes"

DEF_BROWSE_COMMAND = "nautilus obex://[%d]"

import os
import gettext
import __builtin__

translation = gettext.translation("blueman", "/usr/local/share/locale", fallback=True)
translation.install(unicode=True)
__builtin__.ngettext = translation.ungettext

if os.path.exists("../apps") and os.path.exists("../data"):
	BIN_DIR = "./"
	ICON_PATH = "../data/icons"
	UI_PATH = "../data/ui"



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.Config import Config
from blueman.gui.Notification import Notification
from blueman.Sdp import *
from blueman.Lib import rfcomm_list
from blueman.main.SignalTracker import SignalTracker
from blueman.main.Device import Device
from subprocess import PIPE
import dbus
import atexit

import blueman.bluez as Bluez

import gobject
import gtk

class SerialManager(AppletPlugin):
	__icon__ = "blueman-serial"
	__author__ = "Walmis"
	__description__ = _("Standard SPP profile connection handler, allows executing custom actions")
	__author__ = "walmis"
	__options__  = {
		"script": {"type": str, "default": "", 
					"name": _("Script to execute on connection"), 
					"desc": _("<span size=\"small\">The following arguments will be passed:\n"
							"Address, Name, service name, uuid16s, rfcomm node\n"
							"For example:\n"
							"AA:BB:CC:DD:EE:FF, Phone, DUN service, 0x1103, /dev/rfcomm0\n"
							"uuid16s are returned as a comma seperated list\n\n"
							"Upon device disconnection the script will be sent a HUP signal</span>")},
	}
	
	def on_load(self, applet):
		self.signals = SignalTracker()
		self.signals.Handle("dbus", 
							dbus.SystemBus(), 
							self.on_device_property_changed, 
							"PropertyChanged", 
							"org.bluez.Device", 
							path_keyword="path")
		
		self.scripts = {}
		
	def on_unload(self):
		self.signals.DisconnectAll()
		for k in self.scripts.iterkeys():
			self.terminate_all_scripts(k)
		
	def on_device_property_changed(self, key, value, path):
		if key == "Connected" and not value:
			d = Device(path)
			self.terminate_all_scripts(d.Address)
		
		
	def on_rfcomm_connected(self, device, port, uuid):
		uuid16 = sdp_get_serial_type(device.Address, uuid)
		if SERIAL_PORT_SVCLASS_ID in uuid16:
			Notification(_("Serial port connected"), 
			_("Serial port service on device <b>%s</b> now will be available via <b>%s</b>") % (device.Alias, port), 
			pixbuf=get_icon("blueman-serial", 48), 
			status_icon=self.Applet.Plugins.StatusIcon)
			
			self.call_script(device.Address,
				    	 device.Alias,
				   	 sdp_get_serial_name(device.Address, uuid),
				    	 uuid16,
				    	 port)
				    	 
	def terminate_all_scripts(self, address):
		try:
			for p in self.scripts[address].itervalues():
				dprint("Sending HUP to", p.pid)
				os.killpg(p.pid, signal.SIGHUP)
		except:
			pass		
	
	def on_script_closed(self, pid, cond, (address,node)):
		del self.scripts[address][node]
		dprint("Script with PID", pid, "closed")
				    	 
	def manage_script(self, address, node, process):
		if not address in self.scripts:
			self.scripts[address] = {}
			
		if node in self.scripts[address]:
			self.scripts[address][node].terminate()
			
		self.scripts[address][node] = process
		gobject.child_watch_add(process.pid, self.on_script_closed, (address, node))
			
	def call_script(self, address, name, sv_name, uuid16, node):
		c = self.get_option("script")
		if c and c != "":
			args = c.split(" ")
			try:
				args += [address, name, sv_name, ",".join(map(lambda x: hex(x), uuid16)), node]
				dprint(args)
				p = spawn(args, True, reap=False, preexec_fn=lambda: os.setpgid(0, 0))
								
				self.manage_script(address, node, p)
				
			
			except Exception, e:
				Notification(_("Serial port connection script failed"), 
				_("There was a problem launching script %s\n"
				"%s") % (c, str(e)), 
				pixbuf=get_icon("blueman-serial", 48), 
				status_icon=self.Applet.Plugins.StatusIcon)	
				
	def on_rfcomm_disconnect(self, node):
		for k, v in self.scripts.iteritems():
			if node in v:
				dprint("Sending HUP to", v[node].pid)
				os.killpg(v[node].pid, signal.SIGHUP)

			
	def rfcomm_connect_handler(self, device, uuid, reply, err):
		uuid16 = sdp_get_serial_type(device.Address, uuid)
		
		if SERIAL_PORT_SVCLASS_ID in uuid16:
			device.Services["serial"].Connect(uuid, reply_handler=reply, error_handler=err)
		
			return True
		else:
			return False
			
	def on_device_disconnect(self, device):
		self.terminate_all_scripts(device.Address)
		
		if "serial" in device.Services:
			ports = rfcomm_list()
		
			def flt(dev):
				if dev["dst"] == device.Address and dev["state"] == "connected":
					return dev["id"]
		
			active_ports = map(flt, ports)
			
			serial = device.Services["serial"]

			for port in active_ports:
				name = "/dev/rfcomm%d" % port
				try:
					dprint("Disconnecting", name)
					serial.Disconnect(name)
				except:
					dprint("Failed to disconnect", name)
					

@atexit.register
def exit_cleanup():
	if SerialManager.__instance__:
		self = SerialManager.__instance__
		
		for k in self.scripts.iterkeys():
			self.terminate_all_scripts(k)	
		



# BaseInterface.py - abstraction of BlueZ base interfaces
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus
import xml.dom.minidom

from BlueZInterface import BlueZInterface
from ServiceInterface import ServiceInterface

class BaseInterface(BlueZInterface):

    def __init__(self, interface_name, obj_path):
        super(BaseInterface, self).__init__(interface_name, obj_path)
    # __init__

    def ListServiceInterfaces(self):
        '''
        Returns a list of available service interfaces under current object path,
        for available methods for each interface, check BlueZ4 documents.
        '''
        interfaces = []
        dbus_object = dbus.SystemBus().get_object('org.bluez', self.GetObjectPath())
        dbus_introspect = dbus.Interface(dbus_object, 'org.freedesktop.DBus.Introspectable')
        introspect_xml = dbus_introspect.Introspect()
        root_node = xml.dom.minidom.parseString(introspect_xml)
        for interface in root_node.getElementsByTagName('interface'):
            interface_name = interface.getAttribute('name')
            if interface_name != self.GetInterfaceName():
                methods = []
                for method in interface.getElementsByTagName('method'):
                    methods.append(method.getAttribute('name'))
                interfaces.append(ServiceInterface(interface_name, self.GetObjectPath(), methods))
        return interfaces
    # ListServiceInterfaces
# BaseInterface



# Manager.py - class Manager
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import Adapter
import Agent
import errors
from utils import raise_dbus_error
from BaseInterface import BaseInterface

class Manager(BaseInterface):

    '''
    Represents the BlueZ dbus API on interface "org.bluez.Manager".
    Pass the name of the mainloop as a string to __init__.
    The mainloop support could not be changed after the first init of Manager.
    Supported mainloops are gobject, pyqt4 and ecore, use your perferred one.
    And gobject is supported by dbus-python natively.
    If you wanna use pyqt4 or ecore, make sure you have python bindings for them.
    '''

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Manager, cls).__new__(cls)
        return cls._instance
    # __new__

    @raise_dbus_error
    def __init__(self, mainloop):
        try:
            if self.__mainloop_support:
                pass
        except AttributeError:
            self.__setup_event_loop(mainloop)
        super(Manager, self).__init__('org.bluez.Manager', '/')
    # __init__

    def __setup_event_loop(self, mainloop):
        def parse_gobject_mainloop(mainloop):
            try:
                import dbus.mainloop.glib
            except ImportError:
                raise errors.DBusMainLoopModuleNotFoundError('Can not find module for ' + mainloop)
            dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
            return True

        def parse_pyqt4_mainloop(mainloop):
            try:
                import dbus.mainloop.qt
            except ImportError:
                raise errors.DBusMainLoopModuleNotFoundError('Can not find module for ' + mainloop)
            dbus.mainloop.qt.DBusQtMainLoop(set_as_default=True)
            return True

        def parse_ecore_mainloop(mainloop):
            try:
                import e_dbus
            except ImportError:
                raise errors.DBusMainLoopModuleNotFoundError('Can not find module for ' + mainloop)
            e_dbus.DBusEcoreMainLoop(set_as_default=True)
            return True

        _supported_mainloops = {'gobject':parse_gobject_mainloop,
                                'pyqt4':parse_pyqt4_mainloop,
                                'ecore':parse_ecore_mainloop}
        if (not isinstance(mainloop, str)) or (not _supported_mainloops.has_key(mainloop)):
            raise errors.DBusMainLoopNotSupportedError('Supported mainloops are gobject, pyqt4 and ecore')
        else:
            parse_mainloop = _supported_mainloops.get(mainloop)
            if parse_mainloop(mainloop):
                self.__mainloop_support = True
    # __setup_event_loop

    @raise_dbus_error
    def GetAdapter(self, pattern=None):
        '''
        Returns an instance of Adaper for specified adapter,
        valid patterns are "hci0" or "00:11:22:33:44:55".
        Or returns default BlueZ if pattern is None.
        '''
        if pattern is not None:
            obj_path = self.GetInterface().FindAdapter(pattern)
        else:
            obj_path = self.GetInterface().DefaultAdapter()
        return Adapter.Adapter(obj_path)
    # GetAdapter

    @raise_dbus_error
    def ListAdapters(self):
        '''Returns a list of Adapter instances.'''
        obj_paths = self.GetInterface().ListAdapters()
        adapters = []
        for obj_path in obj_paths:
            adapters.append(Adapter.Adapter(obj_path))
        return adapters
    # ListAdapters

    def CreateAgent(self, cls=Agent.Agent, obj_path='/org/bluez/Agent'):
        '''
        Paramater cls should be a custom sub-class of Agent.
        Paramater obj_path is the dbus object path for the agent, should start with '/'.
        Returns an instance of specified cls.
        '''
        if not issubclass(cls, Agent.Agent):
            raise TypeError('Expecting a subclass of Agent')
        return cls(obj_path)
    # CreateAgent
# Manager



# BlueZInterface.py - the base class of other BlueZ interface classes
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus
import types

from utils import raise_type_error

class BlueZInterface(object):

    '''the base class of other BlueZ interface classes.'''

    def __init__(self, interface_name, obj_path):
        self.__obj_path = obj_path
        self.__interface_name = interface_name
        self.__bus = dbus.SystemBus()
        self.__dbus_proxy = self.__bus.get_object('org.bluez', obj_path, follow_name_owner_changes=True)
        self.__interface = dbus.Interface(self.__dbus_proxy, interface_name)
    # __init__

    def GetObjectPath(self):
        return self.__obj_path
    # GetObjectPath

    def GetInterface(self):
        return self.__interface
    # GetInterface

    def GetInterfaceName(self):
        return self.__interface_name
    # GetInterfaceName

    def HandleSignal(self, handler, signal, **kwargs):
        '''
        The handler function will be called when specific signal is emmited.
        For available signals of each interface, check BlueZ4 documents.
        '''
       # raise_type_error(handler, types.FunctionType)
        raise_type_error(signal, types.StringType)
        self.__bus.add_signal_receiver(handler,
                                       signal,
                                       self.__interface_name,
                                       'org.bluez',
                                       self.__obj_path, **kwargs)
    def UnHandleSignal(self, handler, signal, **kwargs):
        '''
        The handler function will be called when specific signal is emmited.
        For available signals of each interface, check BlueZ4 documents.
        '''
       # raise_type_error(handler, types.FunctionType)
        raise_type_error(signal, types.StringType)
        self.__bus.remove_signal_receiver(handler,
                                       signal,
                                       self.__interface_name,
                                       'org.bluez',
                                       self.__obj_path, **kwargs)
    # HandleSignal
# BlueZInterface



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import dbus
from blueman.Sdp import *
from blueman.plugins.AppletPlugin import AppletPlugin
from uuid import uuid1
import gconf
import os
from blueman.main.SignalTracker import SignalTracker
from blueman.main.Device import Device

#class NMApplet

class NMDeviceState:
	UNKNOWN = 0

	#/* Initial state of all devices and the only state for devices not
	# * managed by NetworkManager.
	# *
	# * Allowed next states:
	# *   UNAVAILABLE:  the device is now managed by NetworkManager
	# */
	UNMANAGED = 1

	#/* Indicates the device is not yet ready for use, but is managed by
	# * NetworkManager.  For Ethernet devices, the device may not have an
	# * active carrier.  For WiFi devices, the device may not have it's radio
	# * enabled.
	# *
	# * Allowed next states:
	# *   UNMANAGED:  the device is no longer managed by NetworkManager
	# *   DISCONNECTED:  the device is now ready for use
	# */
	UNAVAILABLE = 2

	#/* Indicates the device does not have an activate connection to anything.
	# *
	# * Allowed next states:
	# *   UNMANAGED:  the device is no longer managed by NetworkManager
	# *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
	# *   PREPARE:  the device has started activation
	# */
	DISCONNECTED = 3

	#/* Indicate states in device activation.
	# *
	# * Allowed next states:
	# *   UNMANAGED:  the device is no longer managed by NetworkManager
	# *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
	# *   FAILED:  an error ocurred during activation
	# *   NEED_AUTH:  authentication/secrets are needed
	# *   ACTIVATED:  (IP_CONFIG only) activation was successful
	# *   DISCONNECTED:  the device's connection is no longer valid, or NetworkManager went to sleep
	# */
	PREPARE = 4
	CONFIG = 5
	NEED_AUTH = 6
	IP_CONFIG = 7

	#/* Indicates the device is part of an active network connection.
	# *
	# * Allowed next states:
	# *   UNMANAGED:  the device is no longer managed by NetworkManager
	# *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
	# *   FAILED:  a DHCP lease was not renewed, or another error
	# *   DISCONNECTED:  the device's connection is no longer valid, or NetworkManager went to sleep
	# */
	ACTIVATED = 8

	#/* Indicates the device's activation failed.
	# *
	# * Allowed next states:
	# *   UNMANAGED:  the device is no longer managed by NetworkManager
	# *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
	# *   DISCONNECTED:  the device's connection is ready for activation, or NetworkManager went to sleep
	# */
	FAILED = 9

class NewConnectionBuilder:
	def __init__(self, parent, params, ok_cb, err_cb):
		self.parent = parent
		self.params = params
		self.ok_cb = ok_cb
		self.err_cb = err_cb
		
		self.signals = SignalTracker()
		
		self.device = None
		self.connection = None
		
		self.signals.Handle("dbus", parent.bus, self.on_nm_device_added, "DeviceAdded", "org.freedesktop.NetworkManager")
		self.signals.Handle("dbus", parent.bus, self.on_nma_new_connection, "NewConnection", "org.freedesktop.NetworkManagerSettings")
		
		self.device = self.parent.find_device(params["bluetooth"]["bdaddr"])

		self.connection = self.parent.find_connection(params["bluetooth"]["bdaddr"], "panu")
		if not self.connection:
			parent.add_connection(params)
			gobject.timeout_add(1000, self.signal_wait_timeout)
		else:
			self.init_connection()
			
	def cleanup(self):
		self.signals.DisconnectAll()
			
	def signal_wait_timeout(self):
		if not self.device or not self.connection:
			self.err_cb(dbus.DBusException("Network Manager did not support the connection"))
			if self.connection:
				self.remove_connection()		
			self.cleanup()
		
	def on_nm_device_added(self, path):
		dprint(path)
		self.device = path
		if self.device and self.connection:
			self.init_connection()
	
	def on_nma_new_connection(self, path):
		dprint(path)
		self.connection = path
		if self.device and self.connection:
			self.init_connection()
			
	def init_connection(self):
		self.cleanup()
		dprint("activating", self.connection, self.device)
		if not self.device or not self.connection:
			self.err_cb(dbus.DBusException("Network Manager did not support the connection"))
			if self.connection:
				self.remove_connection()
			self.cleanup()
		else:
			self.signals.Handle("dbus", self.parent.bus, self.on_device_state, "StateChanged", "org.freedesktop.NetworkManager.Device", path=self.device)
		
			self.parent.nm.ActivateConnection("org.freedesktop.NetworkManagerUserSettings", self.connection, self.device, self.connection)
		
	def remove_connection(self):
		self.parent.remove_connection(self.connection)
		
	def on_device_state(self, state, oldstate, reason):
		dprint("state=",state, "oldstate=", oldstate, "reason=", reason)
		if state <= NMDeviceState.DISCONNECTED and NMDeviceState.DISCONNECTED < oldstate <= NMDeviceState.ACTIVATED:
			if self.err_cb:
				self.err_cb(dbus.DBusException("Connection was interrupted"))
			
			self.remove_connection()
			self.cleanup()
		
		elif state == NMDeviceState.FAILED:
			self.err_cb(dbus.DBusException("Network Manager Failed to activate the connection"))
			self.remove_connection()
			self.cleanup()		
		
		elif state == NMDeviceState.ACTIVATED:
			self.ok_cb()
			self.err_cb = None
			self.ok_cb = None
			
	
class NMPANSupport(AppletPlugin):
	__depends__ = ["DBusService"]
	__conflicts__ = ["DhcpClient", "NMIntegration"]
	__icon__ = "network"
	__author__ = "Walmis"
	__description__ = _("Provides support for Personal Area Networking (PAN) introduced in NetworkManager 0.8")
	__priority__ = 2
	
	def on_load(self, applet):
		self.bus = dbus.SystemBus()
		self.nma = None
		self.nm = None
		self.nm_signals = SignalTracker()
		self.nma_signals = SignalTracker()
		
		self.watch1 = self.bus.watch_name_owner("org.freedesktop.NetworkManagerUserSettings", self.on_nma_owner_changed)
		self.watch2 = self.bus.watch_name_owner("org.freedesktop.NetworkManager", self.on_nm_owner_changed)
		
		self.client = gconf.client_get_default ()
		
	def set_gconf(self, key, value):
		func = None
	
		if type(value) == str or type(value) == unicode:
			func = self.client.set_string
		elif type(value) == int:
			func = self.client.set_int
		elif type(value) == bool:
			func = self.client.set_bool
		elif type(value) == float:
			func = self.client.set_float
		elif type(value) == list:
			def x(key, val):
				self.client.set_list(key, gconf.VALUE_STRING, val)
			func = x
			
		elif type(value) == dbus.Array:
			if value.signature == "i":
				def x(key, val):
					self.client.set_list(key, gconf.VALUE_INT, val)
				func = x
			elif value.signature == "s":
				def x(key, val):
					self.client.set_list(key, gconf.VALUE_STRING, val)
				func = x
			else:
				raise AttributeError("Cant set this type in gconf")							
			
		else:
			raise AttributeError("Cant set %s in gconf" % type(value))
			
		func(key, value)
	
	def find_free_gconf_slot(self):
		dirs = list(self.client.all_dirs ("/system/networking/connections"))
		dirs.sort()
		
		i = 1
		for d in dirs:
			try:
				d = int(os.path.basename(d))
			except:
				continue
			if d != i:
				return i
			
			i+=1
			
		return i
		
	def add_connection(self, params):
		slot = self.find_free_gconf_slot()
		
		base_path = "/system/networking/connections/%d" % slot
		
		for group, settings in params.iteritems():
			path = base_path + "/%s" % group
			for k, v in settings.iteritems():
				key = path + "/%s" % k
				self.set_gconf(key, v)
				
	def remove_connection(self, path):
		self.bus.call_blocking("org.freedesktop.NetworkManagerUserSettings", path, "org.freedesktop.NetworkManagerSettings.Connection", "Delete", "", [])	
				
	def format_bdaddr(self, addr):
		return "%02X:%02X:%02X:%02X:%02X:%02X" % (addr[0], addr[1], addr[2], addr[3], addr[4], addr[5])

	
	def find_device(self, bdaddr):
		devices = self.nm.GetDevices()
		for dev in devices:
			try:
				d = self.bus.call_blocking("org.freedesktop.NetworkManager", dev, "org.freedesktop.DBus.Properties", "GetAll", "s", ["org.freedesktop.NetworkManager.Device.Bluetooth"])	
				if d["HwAddress"] == bdaddr:
					dprint(d["HwAddress"])
					return dev
						
			except dbus.DBusException:
				pass	
							
	def find_connection(self, address, t):
		conns = self.nma.ListConnections()
		for conn in conns:
			c = self.bus.call_blocking("org.freedesktop.NetworkManagerUserSettings", conn, "org.freedesktop.NetworkManagerSettings.Connection", "GetSettings", "", [])
			try:
				if (self.format_bdaddr(c["bluetooth"]["bdaddr"]) == address) and c["bluetooth"]["type"] == t:
					return conn
			except:
				pass
				
	def find_active_connection(self, address, type):
		props = self.bus.call_blocking("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager", "org.freedesktop.DBus.Properties", "GetAll", "s", ["org.freedesktop.NetworkManager"])			
		
		nma_connection = self.find_connection(address, type)
		if nma_connection:
			active_conns = props["ActiveConnections"]
			for conn in active_conns:
				conn_props = self.bus.call_blocking("org.freedesktop.NetworkManager", 
								    conn, 
								    "org.freedesktop.DBus.Properties", 
								    "GetAll", 
								    "s", 
								    ["org.freedesktop.NetworkManager.Connection.Active"])	
								    
				if conn_props["Connection"] == nma_connection:
					return conn	
				
	def on_nma_owner_changed(self, owner):
		if owner == "":
			self.nma = None
		else:
			service = self.bus.get_object("org.freedesktop.NetworkManagerUserSettings", "/org/freedesktop/NetworkManagerSettings")
			self.nma = dbus.proxies.Interface(service, "org.freedesktop.NetworkManagerSettings")		
			
	def on_nm_owner_changed(self, owner):
		if owner == "":
			self.nm = None
			self.nm_signals.DisconnectAll()
		else:		
			service = self.bus.get_object("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager")
			self.nm = dbus.proxies.Interface(service, "org.freedesktop.NetworkManager")
			
		
	def on_unload(self):
		self.nm_signals.DisconnectAll()
		self.nma_signals.DisconnectAll()
		
		self.watch1.cancel()
		self.watch2.cancel()
		
	def service_connect_handler(self, interface, object_path, method, args, ok, err):
		if interface == "org.bluez.Network" and method == "Connect":
			uuid = args[0]
			name = uuid16_to_name(uuid128_to_uuid16(uuid))
			d = Device(object_path)
			
			conn = self.find_active_connection(d.Address, "panu")
			if conn:
				err(dbus.DBusException(_("Already connected")))
			else:
				params = {}
				params["bluetooth"] = {"name": "bluetooth", "bdaddr":  str(d.Address), "type" : "panu"}
				params["connection"] = {"autoconnect": False, "id": str("%s on %s") % (name, d.Alias), "uuid" : str(uuid1()), "type": "bluetooth"}
				params['ipv4'] = {'addresses': dbus.Array([], dbus.Signature("i")), 'dns': dbus.Array([], dbus.Signature("i")), "method": "auto", "routes": dbus.Array([], dbus.Signature("i"))}		
			
				NewConnectionBuilder(self, params, ok, err)
				
			return True
			
		elif interface == "org.bluez.Network" and method == "Disconnect":
			d = Device(object_path)
			active_conn_path = self.find_active_connection(d.Address, "panu")
			if active_conn_path:
				self.bus.call_blocking("org.freedesktop.NetworkManager", 
						       "/org/freedesktop/NetworkManager", 
						       "org.freedesktop.NetworkManager", 
						       "DeactivateConnection", 
						       "o", 
						       [active_conn_path])
				ok()		       	
				return True




# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.Constants import *

try:
    import gtk
except:
    pass
import re
import os
import signal
import atexit
import sys
from subprocess import Popen, call
import subprocess
import commands
import gobject
import traceback
from blueman.Lib import sn_launcher
import __builtin__

GREEN = lambda (x): "\x1b[32;01m" + x + "\x1b[39;49;00m"
BLUE = lambda (x): "\x1b[34;01m" + x + "\x1b[39;49;00m"
BOLD = lambda (x): "\033[1m" + x + "\033[0m"
YELLOW = lambda (x): "\x1b[33;01m" + x + "\x1b[39;49;00m"

import fcntl, struct, termios

try:
    in_fg = os.getpgrp() == struct.unpack('h', fcntl.ioctl(0, termios.TIOCGPGRP, "  "))[0]
except:
    in_fg = False


def dprint(*args):
    #dont print if in the background
    if in_fg:

        s = ""
        for a in args:
            s += (str(a) + " ")
        co = sys._getframe(1).f_code

        fname = BOLD(co.co_name)

        print "_________"
        print "%s %s" % (fname, "(%s:%d)" % (co.co_filename, co.co_firstlineno))
        print s
        try:
            sys.stdout.flush()
        except IOError:
            pass


__builtin__.dprint = dprint

from blueman.main.AppletService import AppletService


def check_bluetooth_status(message, exitfunc, *args, **kwargs):
    try:
        applet = AppletService()
    except:
        print "Blueman applet needs to be running"
        exitfunc()
    if "PowerManager" in applet.QueryPlugins():
        if not applet.GetBluetoothStatus():

            d = gtk.MessageDialog(None, type=gtk.MESSAGE_ERROR)
            d.props.icon_name = "blueman"
            d.props.text = _("Bluetooth Turned Off")
            d.props.secondary_text = message

            d.add_button(gtk.STOCK_QUIT, gtk.RESPONSE_NO)
            d.add_button(_("Enable Bluetooth"), gtk.RESPONSE_YES)
            resp = d.run()
            d.destroy()
            if resp != gtk.RESPONSE_YES:
                exitfunc()
            else:
                applet.SetBluetoothStatus(True, *args, **kwargs)


def wait_for_adapter(bluez_adapter, callback, timeout=1000):
    def on_prop_change(key, value):
        if key == "Powered" and value:
            gobject.source_remove(source)
            bluez_adapter.UnHandleSignal(on_prop_change, "PropertyChanged")
            callback()

    def on_timeout():
        bluez_adapter.UnHandleSignal(on_prop_change, "PropertyChanged")
        gobject.source_remove(source)
        dprint(YELLOW("Warning:"),
               "Bluez didn't provide 'Powered' property in a reasonable timeout\nAssuming adapter is ready")
        callback()

    props = bluez_adapter.GetProperties()
    if props["Address"] != "00:00:00:00:00:00":
        callback()
        return

    source = gobject.timeout_add(timeout, on_timeout)
    bluez_adapter.HandleSignal(on_prop_change, "PropertyChanged")


def startup_notification(name, desc=None, bin_name=None, icon=None):
    dpy = gtk.gdk.display_get_default()
    screen = dpy.get_default_screen().get_number()
    sn = sn_launcher(dpy, screen)
    sn.set_name(name)

    if bin_name:
        sn.set_binary_name(bin_name)
    if icon:
        sn.set_icon_name(icon)

    if desc:
        sn.set_description(desc)

    sn.initiate("", "", gtk.get_current_event_time())

    return sn


def enable_rgba_colormap():
    #screen = gtk.gdk.display_get_default().get_default_screen()
    #colormap = screen.get_rgba_colormap()
    #if colormap == None:
    #	colormap = screen.get_rgb_colormap()
    #gtk.widget_set_default_colormap(colormap)
    pass


def spawn(command, system=False, sn=None, reap=True, *args, **kwargs):
    def child_closed(pid, cond):
        dprint(command, "closed")
        if sn:
            sn.complete()

    if not system:
        if type(command) == list:
            command[0] = os.path.join(BIN_DIR, command[0])
        else:
            command = os.path.join(BIN_DIR, command)
    else:
        if type(command) == list:
            command[0] = os.path.expanduser(command[0])
        else:
            command = os.path.expanduser(command)

    env = os.environ

    if sn:
        id = sn.get_startup_id()
        env["DESKTOP_STARTUP_ID"] = id

    env["BLUEMAN_EVENT_TIME"] = str(gtk.get_current_event_time())

    p = Popen(command, env=env, *args, **kwargs)
    if reap:
        gobject.child_watch_add(p.pid, child_closed)
    return p


def setup_icon_path():
    ic = gtk.icon_theme_get_default()
    ic.prepend_search_path(ICON_PATH)


def get_icon(name, size=24, fallback="gtk-missing-image"):
    ic = gtk.icon_theme_get_default()

    try:
        icon = ic.load_icon(name, size, 0)
    except:
        if not fallback:
            raise
        try:
            icon = ic.load_icon(fallback, size, 0)
        except:
            icon = ic.load_icon("gtk-missing-image", size, 0)

    if icon.props.width > size:
        new_w = size
        new_h = int(size * ( float(icon.props.width) / icon.props.height ))
        icon = icon.scale_simple(new_w, new_h, gtk.gdk.INTERP_BILINEAR)

    if icon.props.height > size:
        new_w = int(size * ( float(icon.props.height) / icon.props.width ))
        new_h = size
        icon = icon.scale_simple(dest_width, dest_height, gtk.gdk.INTERP_BILINEAR)

    return icon


def get_notification_icon(icon, main_icon="blueman"):
    main = get_icon(main_icon, 48)
    sub = get_icon(icon, 24)

    return composite_icon(main, [(sub, 24, 24, 255)])


def adapter_path_to_name(path):
    return re.search(".*(hci[0-9]*)", path).groups(0)[0]

#format error
def e_(msg):
    if isinstance(msg, Exception):
        return str(msg) + "\n" + traceback.format_exc()
    else:
        msg = str(msg)

        s = msg.split(": ")
        del s[0]
        return ": ".join(s)


def opacify_pixbuf(pixbuf, alpha):
    new = pixbuf.copy()
    new.fill(0x00000000)
    pixbuf.composite(new, 0, 0, pixbuf.props.width, pixbuf.props.height, 0, 0, 1, 1, gtk.gdk.INTERP_BILINEAR, alpha)
    return new

#pixbuf, [(pixbuf, x, y, alpha), (pixbuf, x, y, alpha)]

def composite_icon(target, sources):
    target = target.copy()
    for source in sources:
        source[0].composite(target, source[1], source[2], source[0].get_width(), source[0].get_height(), source[1],
                            source[2], 1, 1, gtk.gdk.INTERP_NEAREST, source[3])

    return target


def format_bytes(size):
    ret = 0.0
    size = float(size)
    suffix = ""
    if size < 1024:
        ret = size
        suffix = "B"
    elif size > 1024 and size < (1024 * 1024):
        ret = size / 1024
        suffix = "KB"
    elif size > (1024 * 1024) and size < (1024 * 1024 * 1024):
        ret = size / (1024 * 1024)
        suffix = "MB"
    else:
        ret = size / (1024 * 1024 * 1024)
        suffix = "GB"

    return (ret, suffix)


def create_menuitem(text, image):
    item = gtk.ImageMenuItem(text)
    item.set_image(gtk.image_new_from_pixbuf(image))

    return item


def check_single_instance(id, unhide_func=None):
    print "%s version %s starting" % (id, VERSION)
    cachedir = os.path.expanduser("~/.cache/")
    if not os.path.exists(cachedir):
        os.mkdir(cachedir)
    lockfile = os.path.join(cachedir, "%s-%s" % (id, os.getuid()))

    def handler(signum, frame):
        if unhide_func:
            f = open(lockfile)
            f.readline()
            try:
                event_time = long(f.readline())
            except:
                event_time = 0
            f.close()
            unhide_func(event_time)


    signal.signal(signal.SIGUSR1, handler)

    if os.path.exists(lockfile):
        f = open(lockfile)
        try:
            pid = int(f.readline())
        except:
            pid = 0

        try:
            event_time = int(f.readline())
        except:
            event_time = 0
        f.close()
        if pid > 0:
            isrunning = os.path.exists("/proc/%s" % pid)
            if isrunning:
                try:
                    f = open("/proc/%s/cmdline" % pid)
                    cmdline = f.readline().replace("\0", " ")
                    f.close()
                except:
                    cmdline = None
            if not isrunning or (isrunning and id not in cmdline):
                print "Stale PID, overwriting"
                os.remove(lockfile)
            else:
                print "There is an instance already running"
                time = os.getenv("BLUEMAN_EVENT_TIME") or 0

                f = file(lockfile, "w")
                f.write("%s\n%s" % (str(pid), str(time)))
                f.close()

                os.kill(pid, signal.SIGUSR1)
                exit()
        else:
            os.remove(lockfile)

    try:
        fd = os.open(lockfile, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0664)
        os.write(fd, "%s\n%s" % (str(os.getpid()), "0"))
        os.close(fd)
    except OSError:
        print "There is an instance already running"
        exit()

    atexit.register(lambda: os.remove(lockfile))


def have(t):
    out = call(["which", t], stdout=subprocess.PIPE)

    return out != 1


def mask_ip4_address(ip, subnet):
    masked_ip = ""
    for i in range(4):
        masked_ip += chr(ord(ip[i]) & ord(subnet[i]))

    return masked_ip



# Agent.py - class Agent and decorator AgentMethod
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import inspect
import dbus
import dbus.service
import errors

__SIGNATURES__ = {'Release':('', ''),
                  'RequestPinCode':('o', 's'),
                  'RequestPasskey':('o', 'u'),
                  'DisplayPasskey':('ouu', ''),
                  'RequestConfirmation':('ou', ''),
                  'Authorize':('os', ''),
                  'ConfirmModeChange':('s', ''),
                  'Cancel':('', '')}

def AgentMethod(func):
    '''
    The decorator for customizing the agent methods.
    To use async callbacks, add two extra parameters for
    success callback and error callback in the def of the agent method.
    '''
    global __SIGNATURES__
    try:
        signatures = __SIGNATURES__[func.__name__]
    except KeyError:
        raise errors.BluezUnavailableAgentMethodError('method name ' + func.__name__ + ' unavailable for agent')
    args = inspect.getargspec(func)[0]
    if len(args) - len(signatures[0]) == 3:
        async_callbacks = (args[-2], args[-1])
    else:
        async_callbacks = None
    warp = dbus.service.method('org.bluez.Agent',
                               in_signature=signatures[0],
                               out_signature=signatures[1],
                               async_callbacks=async_callbacks)
    return warp(func)
# AgentMethod

class Agent(dbus.service.Object):

    '''
    Represents the BlueZ dbus API on interface "org.bluez.Agent".
    Inherit from this class and use AgentMethod decorator
    to customize the methods.
    The simple-agent is provided by default.
    '''

    def __init__(self, obj_path):
        self.__obj_path = obj_path
        dbus.service.Object.__init__(self, dbus.SystemBus(), obj_path)
    # __init__

    def GetObjectPath(self):
        '''Returns the dbus object path of the agent.'''
        return self.__obj_path
    # GetObjectPath

    @AgentMethod
    def Release(self):
        '''
        This method gets called when the service daemon
        unregisters the agent. An agent can use it to do
        cleanup tasks. There is no need to unregister the
        agent, because when this method gets called it has
        already been unregistered.
        '''
        dprint("Release")
    # Release

    @AgentMethod
    def RequestPinCode(self, device):
        '''
        This method gets called when the service daemon
        needs to get the passkey for an authentication.
        The return value should be a string of 1-16 characters
        length. The string can be alphanumeric.
        '''
        dprint("RequestPinCode (%s)" % (device))

    # RequestPinCode

    @AgentMethod
    def RequestPasskey(self, device):
        '''
        This method gets called when the service daemon
        needs to get the passkey for an authentication.
        The return value should be a numeric value
        between 0-999999.
        '''
        dprint("RequestPasskey (%s)" % (device))

    # RequestPasskey

    @AgentMethod
    def DisplayPasskey(self, device, passkey, entered):
        '''
        This method gets called when the service daemon
        needs to display a passkey for an authentication.
        The entered parameter indicates the number of already
        typed keys on the remote side.
        An empty reply should be returned. When the passkey
        needs no longer to be displayed, the Cancel method
        of the agent will be called.
        During the pairing process this method might be
        called multiple times to update the entered value.
        '''
        dprint("DisplayPasskey (%s, %d)" % (device, passkey))
    # DisplayPasskey

    @AgentMethod
    def RequestConfirmation(self, device, passkey):
        '''
        This method gets called when the service daemon
        needs to confirm a passkey for an authentication.
        To confirm the value it should return an empty reply
        or an error in case the passkey is invalid.
        '''
        dprint("RequestConfirmation (%s, %d)" % (device, passkey))

    # RequestConfirmation

    @AgentMethod
    def Authorize(self, device, uuid):
        '''
        This method gets called when the service daemon
        needs to authorize a connection/service request.
        '''
        dprint("Authorize (%s, %s)" % (device, uuid))
    # Authorize

    @AgentMethod
    def ConfirmModeChange(self, mode):
        '''
        This method gets called if a mode change is requested
        that needs to be confirmed by the user. An example
        would be leaving flight mode.
        '''
        dprint("ConfirmModeChange (%s)" % (mode))
    # ConfirmModeChange

    @AgentMethod
    def Cancel(self):
        '''
        This method gets called to indicate that the agent
        request failed before a reply was returned.
        '''
        dprint("Cancel")
    # Cancel
# Agent



# Device.py - class Device
#
# Copyright (C) 2008 Vinicius Gomes <vcgomes [at] gmail [dot] com>
# Copyright (C) 2008 Li Dongyang <Jerry87905 [at] gmail [dot] com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from utils import raise_dbus_error
from BaseInterface import BaseInterface

class Device(BaseInterface):

    '''Represents the BlueZ dbus API on interface "org.bluez.Device".'''

    @raise_dbus_error
    def __init__(self, obj_path):
        super(Device, self).__init__('org.bluez.Device', obj_path)
    # __init__

    @raise_dbus_error
    def GetProperties(self):
        '''
        Returns all properties for the device.
        Properties:
            string Address [readonly]
                The Bluetooth device address of the remote device.
            string Name [readonly]
                The Bluetooth remote name. This value can not be
                changed. Use the Alias property instead.
            uint32 Class [readonly]
                The Bluetooth class of device of the remote device.
            array{string} UUIDs [readonly]
                List of 128-bit UUIDs that represents the available
                remote services.
            boolean Paired [readonly]
                Indicates if the remote device is paired.
            boolean Connected [readonly]
                Indicates if the remote device is currently connected.
                A PropertyChanged signal indicate changes to this
                status.
            boolean Trusted [readwrite]
                Indicates if the remote is seen as trusted. This
                setting can be changed by the application.
            string Alias [readwrite]
                The name alias for the remote device. The alias can
                be used to have a different friendly name for the
                remote device.
                In case no alias is set, it will return the remote
                device name. Setting an empty string as alias will
                convert it back to the remote device name.
                When reseting the alias with an empty string, the
                emitted PropertyChanged signal will show the remote
                name again.
            object Adapter [readonly]
                The object path of the adpater the device belongs to.
        '''
        return self.GetInterface().GetProperties()
    # GetProperties

    @raise_dbus_error
    def SetProperty(self, name, value):
        '''
        Changes the value of the specified property. Only
        properties that are listed a read-write are changeable.
        On success this will emit a PropertyChanged signal.
        '''
        self.GetInterface().SetProperty(name, value)
    # SetProperty

    @raise_dbus_error
    def DiscoverServices(self, pattern):
        '''
        This method starts the service discovery to retrieve
        remote service records. The pattern parameter can
        be used to specific specific UUIDs.
        The return value is a dictionary with the record
        handles as keys and the service record in XML format
        as values. The key is uint32 and the value a string
        for this dictionary.
        '''
        return self.GetInterface().DiscoverServices(pattern)
    # DiscoverServices

    @raise_dbus_error
    def CancelDiscovery(self):
        '''
        This method will cancel any previous DiscoverServices
        transaction.
        '''
        self.GetInterface().CancelDiscovery()
    # CancelDiscovery

    @raise_dbus_error
    def Disconnect(self, *args, **kwargs):
        '''
        This method disconnects a specific remote device by
        terminating the low-level ACL connection. The use of
        this method should be restricted to administrator
        use.
        A DisconnectRequested signal will be sent and the
        actual disconnection will only happen 2 seconds later.
        This enables upper-level applications to terminate
        their connections gracefully before the ACL connection
        is terminated.
        '''
        self.GetInterface().Disconnect(*args, **kwargs)
    # Disconnect
# Device



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluez.Device import Device
from blueman.gui.Notification import Notification
from blueman.main.Mechanism import Mechanism
from blueman.main.PPPConnection import PPPConnection
from blueman.main.Config import Config

import gobject

from blueman.Sdp import *
import os
import dbus

class connection:
		
	def __init__(self, applet, device, port, ok, err):
		self.reply_handler = ok
		self.error_handler = err
		self.device = device
		self.port = port
		self.Applet = applet
		
		res = os.popen("ps x -o pid,args | grep modem-manager").read()
		if not res:
			self.connect()
		else:
			dprint("ModemManager is running, delaying connection 5sec for it to complete probing")
			gobject.timeout_add(5000, self.connect)
		
	def connect(self):
		c = Config("gsm_settings/" + self.device.Address)
		if c.props.apn == None:
			c.props.apn = ""
			
		if c.props.number == None:
			c.props.number = "*99#"	
		
		m = Mechanism()
		m.PPPConnect(self.port, c.props.number, c.props.apn, reply_handler=self.on_connected, error_handler=self.on_error, timeout=200)	
	
	def on_error(self, error):
		self.error_handler(error)
		gobject.timeout_add(1000, self.device.Services["serial"].Disconnect, self.port)
		
	def on_connected(self, iface):
		self.reply_handler(self.port)
		self.Applet.Plugins.Run("on_ppp_connected", self.device, self.port, iface)
		
		Notification(_("Connected"), _("Successfully connected to <b>DUN</b> service on <b>%(0)s.</b>\nNetwork is now available through <b>%(1)s</b>") % {"0":self.device.Alias, "1":iface}, pixbuf=get_icon("network-wireless", 48), status_icon=self.Applet.Plugins.StatusIcon)	

	

class PPPSupport(AppletPlugin):
	__depends__ = ["DBusService"]
	__conflicts__ = ["NMIntegration"]
	__description__ = _("Provides basic support for connecting to the internet via DUN profile.")
	__author__ = "Walmis"
	__icon__ = "modem"
	__priority__ = 0
	
	def on_load(self, applet):
		AppletPlugin.add_method(self.on_ppp_connected)
		
	def on_unload(self):
		pass
		
	def on_ppp_connected(self, device, rfcomm, ppp_port):
		pass
		
	def on_rfcomm_connected(self, device, port, uuid):
		pass
		
	def rfcomm_connect_handler(self, device, uuid, reply, err):
		uuid16 = sdp_get_serial_type(device.Address, uuid)
		if DIALUP_NET_SVCLASS_ID in uuid16:
		
			def local_reply(port):
				connection(self.Applet, device, port, reply, err)
				#self.port2device[port] = device
					
			device.Services["serial"].Connect(uuid, reply_handler=local_reply, error_handler=err)
			dprint("Connecting rfcomm device")
		
			return True
		else:
			return False
